// Test main.

#include <cassert>
#include <iostream>

using std::cout;
using std::endl;

#include "..\StateMachine.h"

#include "TestActions.h"

// States we might start in.
enum StartingState
{
    START_IN_DAY,
    START_IN_OFF,
    START_IN_MOVING,
    START_IN_TIMING
};

struct TestVector
{
    // Type for a pointer to void void member function of StateMachine. 
    typedef void (StateMachine::* SMFunctionPointer) (void);

    const char *        testTitle;
    StartingState       startingState;
    SMFunctionPointer   eventFunctionToApply;

    const char *        expectedState;
    const TestActions::ActionType *firstAction;
    const TestActions::ActionType *secondAction;
    const TestActions::ActionType *thirdAction;
};

static void TestOneAction (TestActions &actions, const TestActions::ActionType &expected)
{
    if (!actions.v.empty())
    {
        TestActions::ActionType &action = *actions.v.begin();
        assert (action == expected);
        actions.v.erase (actions.v.begin());
    }
    else
    {
        assert ("action list unexpectedly empty");
    }
}

static void TestOneTransition (const TestVector &v)
{
    TestActions resultingActions;
    StateMachine uut (resultingActions);

    // put it in the starting state
    {
        switch (v.startingState)
        {
        case START_IN_DAY: uut.ChangeState (uut.DayState());
            break;
        case START_IN_OFF: uut.ChangeState (uut.OffState());
            break;
        case START_IN_MOVING: uut.ChangeState (uut.MovingState());
            break;
        case START_IN_TIMING: uut.ChangeState (uut.TimingState());
            break;
        default:
            assert ("Unknown starting state");
            return;
        }

        // clear results of changing to the starting state.
        resultingActions.v.clear();

        cout << "Starting state is: " << uut.Report() << endl;
    }

    (uut.*(v.eventFunctionToApply))();

    if (v.firstAction)
    {
        TestOneAction (resultingActions, *v.firstAction);
    }

    if (v.secondAction)
    {
        TestOneAction (resultingActions, *v.secondAction);
    }

    if (v.thirdAction)
    {
        TestOneAction (resultingActions, *v.thirdAction);
    }

    while (!resultingActions.v.empty())
    {
        assert ("Found unexpected action");
        resultingActions.v.erase (resultingActions.v.begin());
    }

    cout << "Resulting state is " << v.expectedState << endl;
    assert (strcmp (uut.Report(), v.expectedState) == 0);
}

int main (void)
{
    TestActions ta;
    StateMachine sm(ta);

    const TestVector v[] =
    {
        { "Day + dark -> off",              START_IN_DAY, StateMachine::Dark,       "Off", 0, 0, 0 },
        { "Day + light -> no change",       START_IN_DAY, StateMachine::Light,      "Day", 0, 0, 0 },
        { "Day + movement -> no change",    START_IN_DAY, StateMachine::Movement,   "Day", 0, 0, 0 },
        { "Day + no_movement -> no change", START_IN_DAY, StateMachine::NoMovement, "Day", 0, 0, 0 },
        { "Day + timeout -> no change",     START_IN_DAY, StateMachine::Timeout,    "Day", 0, 0, 0 },

        { "Off + dark -> no change",           START_IN_OFF, StateMachine::Dark,       "Off", 0, 0, 0 },
        { "Off + light -> lamp off; day",      START_IN_OFF, StateMachine::Light,      "Day",    &LampOff, 0, 0 },
        { "Off + movement -> lamp on; moving", START_IN_OFF, StateMachine::Movement,   "Moving", &LampOn,  0, 0 },
        { "Off + no_movement -> no change",    START_IN_OFF, StateMachine::NoMovement, "Off", 0, 0, 0 },
        { "Off + timeout -> no change",        START_IN_OFF, StateMachine::Timeout,    "Off", 0, 0, 0 },

        { "Moving + dark -> no change",        START_IN_MOVING, StateMachine::Dark,       "Moving", 0, 0, 0 },
        { "Moving + light -> lamp off; day",   START_IN_MOVING, StateMachine::Light,      "Day",    &LampOff, 0, 0 },
        { "Moving + movement -> no change",    START_IN_MOVING, StateMachine::Movement,   "Moving", 0, 0, 0 },
        { "Moving + no_movement -> start timer; timing", START_IN_MOVING, StateMachine::NoMovement, "Timing", &StartTimer, 0, 0 },
        { "Moving + timeout -> no change",     START_IN_MOVING, StateMachine::Timeout,    "Moving", 0, 0, 0 },

        { "Timing + dark -> no change",        START_IN_TIMING, StateMachine::Dark,       "Timing", 0, 0, 0 },
        { "Timing + light -> lamp off; day",   START_IN_TIMING, StateMachine::Light,      "Day",    &LampOff, 0, 0 },
        { "Timing + movement -> moving",       START_IN_TIMING, StateMachine::Movement,   "Moving", 0, 0, 0 },
        { "Timing + no_movement -> no change", START_IN_TIMING, StateMachine::NoMovement, "Timing", &StartTimer, 0, 0 },
        { "Timing + timeout -> lamp off; off", START_IN_TIMING, StateMachine::Timeout,    "Off", &LampOff, 0, 0 },

        { 0, START_IN_DAY, StateMachine::Dark, 0, 0, 0, 0 }
    };

    for (unsigned i = 0; v[i].testTitle; i++)
    {
        TestOneTransition (v[i]);
    }

    return 0;
}
